import torch


def white(length: int, seed: int, device: str = "cpu") -> torch.Tensor:
    gen = torch.Generator(device=device)
    gen.manual_seed(seed)
    return torch.rand(length, generator=gen, device=device) * 2.0 - 1.0


def blue(length: int, sr: float, seed: int, device: str = "cpu") -> torch.Tensor:
    offset = int(length / 2)
    # white noise
    wh = white(length + (offset * 2), seed, device)
    # fft
    WH_f = torch.fft.rfft(wh)
    freqs = torch.fft.rfftfreq(len(wh), 1 / sr)
    # white -> blue
    BL_f = WH_f * torch.sqrt(freqs)
    # irfft
    bl = torch.fft.irfft(BL_f)
    # normalize
    bl /= bl.abs().max()

    return bl[offset : length + offset]


def pink(length: int, sr: float, seed: int, device: str = "cpu") -> torch.Tensor:
    offset = int(length / 2)
    # white noise
    wh = white(length + (offset * 2), seed, device)
    # fft
    WH_f = torch.fft.rfft(wh)
    freqs = torch.fft.rfftfreq(len(wh), 1 / sr)
    # white -> pink
    PK_f = WH_f.clone()
    for i in range(len(WH_f)):
        PK_f[i] = WH_f[i] / torch.sqrt(freqs[i]) if 20 < freqs[i] else 0
    # irfft
    pk = torch.fft.irfft(PK_f)
    # normalize
    pk /= pk.abs().max()

    return pk[offset : length + offset]



from dataclasses import dataclass
from math import inf
from pathlib import Path
from typing import Final

import librosa
import numpy as np
import numpy.random as rand
import numpy.typing as npt

from .color_noise import pink as pk_noise
from .color_noise import white as wh_noise


@dataclass
class _TimeAlignment:
    def __init__(self, start, end, phoneme) -> None:
        self.start: Final[float] = start
        self.end: Final[float] = end
        self.phoneme: Final[str] = phoneme


class WavLabHandler:
    __x: Final[npt.NDArray[np.floating]]
    __sr: Final[float]
    __monophone_label: Final[list[_TimeAlignment]]

    def __init__(self, wav_path: Path, lab_path: Path) -> None:
        # load audio
        self.__x, self.__sr = librosa.load(wav_path, sr=None)
        self.__x.flags.writeable = False  # x is immutable

        # load label
        with lab_path.open() as f:
            self.__monophone_label = []
            for line in f:
                sp: list[str] = line.split()
                align = _TimeAlignment(
                    start=int(sp[0]) / 1e7,
                    end=int(sp[1]) / 1e7,
                    phoneme=sp[2],
                )
                self.__monophone_label.append(align)
        if len(self.__monophone_label) < 3:
            raise ValueError("Invalid label format")

    def get_answer(self) -> tuple[float, float]:
        return (
            self.__monophone_label[1].start,
            self.__monophone_label[-1].start,
        )

    def get_signal(self) -> tuple[npt.NDArray[np.floating], float]:
        return self.__x, self.__sr

    def get_noise_signal(
        self, snr: float, is_white: bool, with_pulse: bool, noise_seed: int
    ) -> tuple[npt.NDArray, int]:
        x: npt.NDArray[np.floating] = self.__x.copy()
        sr: int = int(self.__sr)
        ans_s_sec, ans_e_sec = self.get_answer()
        speech_start_idx: int = int(ans_s_sec * sr)
        speech_end_idx: int = int(ans_e_sec * sr)

        # generate noise (white or pink)
        noise: npt.NDArray[np.floating] = (
            (
                wh_noise(len(x), noise_seed)
                if is_white
                else pk_noise(len(x), sr, noise_seed)
            )
            .cpu()
            .numpy()
        )

        # mix stationary noise and signal (in specified snr)
        if snr == inf:
            noised_x = x
        elif snr == -inf:
            noised_x = noise
        else:
            noise_scale = _determine_noise_scale(
                x[speech_start_idx:speech_end_idx], noise, int(snr)
            )
            noised_x = x + noise * noise_scale

        # add pulse noise
        if with_pulse:
            rand.seed(noise_seed)
            pulse: npt.NDArray[np.floating] = rand.random(2) - 0.5 * 2
            # determine index adding pulse noise
            start_pulse_index: int = np.random.randint(0, speech_start_idx)
            end_pulse_index: int = np.random.randint(speech_end_idx, len(x) - 1)
            # add pulse noise
            noised_x[start_pulse_index] = pulse[0]
            noised_x[end_pulse_index] = pulse[1]
        return noised_x, sr

    def get_noise_signal2(
        self, snr: float, noise_type: str, noise_seed: int
    ) -> tuple[npt.NDArray, float]:
        x: npt.NDArray = self.__x
        ans_s_sec, ans_e_sec = self.get_answer()
        speech_start_idx: Final[int] = int(ans_s_sec * self.__sr)
        speech_end_idx: Final[int] = int(ans_e_sec * self.__sr)

        color_flag = False

        for nt in noise_type.split():
            if not color_flag and (nt == "white" or nt == "pink"):
                # generate color noise
                noise: npt.NDArray[np.floating] = (
                    (
                        wh_noise(len(x), noise_seed)
                        if nt == "white"
                        else pk_noise(len(x), self.__sr, noise_seed)
                    )
                    .cpu()
                    .numpy()
                )

                # mix stationary noise and signal (in specified snr)
                if snr == inf:
                    noised_x = x.copy()
                elif snr == -inf:
                    noised_x = noise
                else:
                    noise_scale = _determine_noise_scale(
                        x[speech_start_idx:speech_end_idx], noise, int(snr)
                    )
                    noised_x = x + noise * noise_scale

                color_flag = True

            # add pulse noise
            if nt == "pulse":
                rand.seed(noise_seed)
                pulse: npt.NDArray[np.floating] = rand.random(2) - 0.5 * 2
                # determine index adding pulse noise
                start_pulse_index: int = np.random.randint(0, speech_start_idx)
                end_pulse_index: int = np.random.randint(speech_end_idx, len(x) - 1)
                # add pulse noise
                noised_x[start_pulse_index] = pulse[0]
                noised_x[end_pulse_index] = pulse[1]

        return noised_x, self.__sr


def _determine_noise_scale(
    signal: npt.NDArray[np.floating],
    noise: npt.NDArray[np.floating],
    desired_snr_db: int,
) -> float:
    desired_snr_linear: float = 10 ** (desired_snr_db / 10)
    signal_power: float = np.mean(signal**2)
    noise_power: float = np.mean(noise**2)
    scaling_factor: float = np.sqrt(signal_power / (desired_snr_linear * noise_power))
    return scaling_factor



# SPDX-License-Identifier:GPL-3.0-or-later

from typing import Final, Optional

import noisereduce as nr
import numpy as np
import numpy.typing as npt
from librosa import resample
from scipy.signal import cheby1, firwin, lfilter, sosfilt

from ..utility import color_noise
from ..utility.constants import F0_CEIL, F0_FLOOR, NYQ, SR
from ..utility.signal import normalize
from ..utility.signal import root_mean_square as rms
from ..utility.signal import slide_index
from ..utility.signal import zero_crossing_rate as zcr


def vsed_debug_v1(
    y: npt.NDArray,
    orig_sr: float,
    # -------------------------------------------
    win_length_s: Optional[float] = None,
    hop_length_s: float = 0.01,
    # -------------------------------------------
    rms_threshold: float = 0.03,
    zcr_threshold: float = 0.67,
    zcr_margin_s: float = 0.1,
    offset_s: float = 0.03,
    # -------------------------------------------
    noise_seed: int = 0,
):
    # resample
    y_rsp: Final[npt.NDArray] = resample(
        y=y, orig_sr=orig_sr, target_sr=SR, res_type="soxr_lq"
    )

    # constants
    win_length_s = win_length_s if win_length_s is not None else hop_length_s * 4
    win_length: Final[int] = int(win_length_s * SR)
    hop_length: Final[int] = int(hop_length_s * SR)
    zcr_margin: Final[int] = int(zcr_margin_s / hop_length_s)

    # preprocess: add blue noise && remove background noise
    y_nr: Final[npt.NDArray] = _00_preprocess(y_rsp, SR, noise_seed)

    # step1: Root mean square
    start1, end1, y_rms = _01_rms(y_nr, SR, rms_threshold, win_length, hop_length)

    # step2: Zero cross
    start2, end2, y_zcr = _02_zcr(
        y_nr,
        SR,
        start1,
        end1,
        zcr_threshold,
        zcr_margin,
        win_length,
        hop_length,
    )

    # index -> second: rms
    start1_s: Final[float] = max(0, start1 * hop_length_s)
    end1_s: Final[float] = min(end1 * hop_length_s, len(y_rsp) / SR)

    # index -> second: zrs
    start2_s: Final[float] = max(0, start2 * hop_length_s)
    end2_s: Final[float] = min(end2 * hop_length_s, len(y_rsp) / SR)

    # add offset
    start3_s: Final[float] = max(0, start2_s - offset_s)
    end3_s: Final[float] = min(end2_s + offset_s, len(y_rsp) / SR)

    feats_timestamp: Final[npt.NDArray] = np.linspace(
        0, len(y_zcr) * hop_length_s, len(y_zcr)
    )

    return (
        start1_s,
        end1_s,
        start2_s,
        end2_s,
        start3_s,
        end3_s,
        feats_timestamp,
        y_rms,
        y_zcr,
    )


def _00_preprocess(y: npt.NDArray, sr: int, noise_seed: int) -> npt.NDArray:
    data_rms: Final[npt.NDArray] = np.sort(rms(y, 2048, 512))  # <- default of librosa
    signal_amp: Final[float] = data_rms[-2]
    noise_amp: Final[float] = max(data_rms[1], 1e-10)
    snr: Final[float] = min(20 * np.log10(signal_amp / noise_amp), 10)
    noise: Final[npt.NDArray] = color_noise.blue(len(y), sr, noise_seed).cpu().numpy()
    y_blue: Final[npt.NDArray] = y + noise * (signal_amp / 10 ** (snr / 20))
    y_blue_normalized: Final[npt.NDArray] = (
        y_blue if np.max(np.abs(y_blue)) <= 1 else y_blue / np.max(np.abs(y_blue))
    )
    return nr.reduce_noise(y_blue_normalized, sr)


def _01_rms(
    y: npt.NDArray, sr: int, threshold: float, win_length: int, hop_length: int
) -> tuple[int, int, npt.NDArray]:
    wp: Final[tuple[float, float]] = (F0_FLOOR / NYQ, F0_CEIL / NYQ)
    band_sos: Final[npt.NDArray] = cheby1(12, 1, wp, "bandpass", output="sos")
    y_bpf: Final[npt.NDArray] = sosfilt(band_sos, y)
    y_rms: Final[npt.NDArray] = normalize(rms(y_bpf, win_length, hop_length))
    start1: Final[int] = (
        np.where(threshold < y_rms)[0][0] if np.any(threshold < y_rms) else 0
    )
    end1: Final[int] = (
        np.where(threshold < y_rms)[0][-1]
        if np.any(threshold < y_rms)
        else len(y_rms) - 1
    )
    return start1, end1, y_rms


def _02_zcr(
    y: npt.NDArray,
    sr: int,
    start1: int,
    end1: int,
    threshold: float,
    margin: int,
    win_length: int,
    hop_length: int,
) -> tuple[int, int, npt.NDArray]:
    high_b: Final[float] = firwin(101, F0_CEIL, pass_zero=False, fs=sr)
    y_hpf: Final[npt.NDArray] = lfilter(high_b, 1.0, y)
    y_zcr: Final[npt.NDArray] = normalize(zcr(y_hpf, win_length, hop_length))
    # slide start index
    start2: Final[int] = slide_index(
        goto_min=True,
        y=y_zcr,
        start_idx=start1,
        threshold=threshold,
        margin=margin,
    )
    # slide end index
    end2: Final[int] = slide_index(
        goto_min=False,
        y=y_zcr,
        start_idx=end1,
        threshold=threshold,
        margin=margin,
    )
    return start2, end2, y_zcr
